package main

import (
	"testing"
)

type IntEdge struct {
	source *IntNode
	target *IntNode
	length int
}

func (ie *IntEdge) Source() Node {
	return ie.source
}

func (ie *IntEdge) Target() Node {
	return ie.target
}

func (ie *IntEdge) Length() int {
	return ie.length
}

func test_conversion() {
	var _ Edge = &IntEdge{}
	var _ Node = &IntNode{}
}

type IntNode struct {
	Val   int
	edges []IntEdge
}

func (in *IntNode) Edges() []Edge {
	var res []Edge
	for i := range in.edges {
		res = append(res, &in.edges[i])
	}
	return res
}

func (in *IntNode) Index() interface{} {
	return in.Val
}

func NewIntNode(v int) *IntNode {
	in := &IntNode{}
	in.Val = v
	return in
}

func TestBFS_Add(t *testing.T) {
	bfs := NewBFS()
	bfs.Add(NewIntNode(1), 0)
	if bfs.queue.Len() != 1 {
		t.Fatalf("1 did not get put into the queue")
	}
	if val := bfs.queue.Front().Value.(*IntNode).Val; val != 1 {
		t.Fatalf("the first entry is not 1: is %v", val)
	}
	if _, present := bfs.distances[1]; !present {
		t.Fatalf("the entry 1 was not properly put into the distances set")
	}
	bfs.Add(NewIntNode(2), 0)
	if bfs.queue.Len() != 2 {
		t.Fatalf("2 did not get put into the queue")
	}
	if bfs.queue.Front().Next().Value.(*IntNode).Val != 2 {
		t.Fatalf("the second entry is not 2")
	}
	if _, present := bfs.distances[2]; !present {
		t.Fatalf("the entry 2 was not properly put into the distances set")
	}
	// should be a no-op
	bfs.Add(NewIntNode(1), 0)
	if bfs.queue.Len() != 2 {
		t.Fatalf("1 was not properly ignored being added a second time")
	}
}

func TestBFS_AddShorterDistance(t *testing.T) {
	bfs := NewBFS()
	bfs.Add(NewIntNode(1), 5)
	bfs.Add(NewIntNode(1), 2)
	if bfs.queue.Len() != 1 {
		t.Fatalf("1 either wasn't in the queue or was added twice")
	}
	if bfs.distances[1] != 2 {
		t.Fatalf("Distance to [1] was not properly updated")
	}
	if n, d := bfs.Next(); n.(*IntNode).Val != 1 || d != 2 {
		t.Fatalf("Node or distance was not correct.")
	}
}

func TestBFS_Next_simple(t *testing.T) {
	bfs := NewBFS()
	bfs.Add(NewIntNode(1), 0)
	bfs.Add(NewIntNode(2), 0)
	if bfs.queue.Len() != 2 || len(bfs.distances) != 2 {
		t.Fatalf("1 and 2 were not properly added to the queue.")
	}
	if first, _ := bfs.Next(); first.(*IntNode).Val != 1 {
		t.Fatalf("1 was not the first entry in the bfs")
	}
	if bfs.queue.Len() != 1 {
		t.Fatalf("1 was not removed from the queue")
	}
	// We remember all entries we've walked, so we don't walk them a second time
	if _, ok := bfs.distances[1]; !ok {
		t.Fatalf("1 was removed from the entries list")
	}
	if second, _ := bfs.Next(); second.(*IntNode).Val != 2 {
		t.Fatalf("2 was not the second entry in the bfs")
	}
	if last, _ := bfs.Next(); last != nil {
		t.Fatalf("we didn't get nil when the queue was empty.")
	}
}

func TestBFS_Next(t *testing.T) {
	bfs := NewBFS()
	node1 := NewIntNode(1)
	node2 := NewIntNode(2)
	node3 := NewIntNode(3)
	node4 := NewIntNode(4)
	edge12 := IntEdge{source: node1, target: node2, length: 1}
	edge13 := IntEdge{source: node1, target: node3, length: 1}
	edge24 := IntEdge{source: node2, target: node4, length: 1}
	edge32 := IntEdge{source: node3, target: node2, length: 1}
	node1.edges = append(node1.edges, edge12, edge13)
	node2.edges = append(node2.edges, edge24)
	node3.edges = append(node3.edges, edge32)
	bfs.Add(node1, 0)
	if n, d := bfs.Next(); n.(*IntNode) != node1 && d != 0 {
		t.Fatalf("Next did not return first node and distance"+
			", returned: %v, %d not %v, %d",
			n, d, node1, 0)
	}
	if bfs.queue.Len() != 2 {
		t.Fatalf("did not add new nodes to the queue")
	}
	if n, d := bfs.Next(); n.(*IntNode) != node2 && d != 1 {
		t.Fatalf("expected the second node to be node2 (%v) not %v",
			node2, n)
	}
	if n, d := bfs.Next(); n.(*IntNode) != node3 && d != 1 {
		t.Fatalf("expected the third node to be node3 (%v) not %v",
			node3, n)
	}
	if n, d := bfs.Next(); n.(*IntNode) != node4 && d != 2 {
		t.Fatalf("expected the fourth node to be node4 (%v) not %v",
			node4, n)
	}
	if n, d := bfs.Next(); n != nil && d != 0 {
		t.Fatalf("expected the final node to be nil not %v",
			n)
	}
}
