package main

import (
	"testing"
)

func TestDenseIntSet_create(t *testing.T) {
	dis := DenseIntSet{}
	if len(dis.bits) != 0 {
		t.Errorf("wrong length, expected %d got %d", 0, len(dis.bits))
	}
}

func assertOffset(t *testing.T, value uint, expected_offset uint, expected_mask uint32) {
	o, m := offset(value)
	if o != expected_offset || m != expected_mask {
		t.Errorf("Incorrect result from offset(). Expected %u %08x, got %u %08x",
			expected_offset, expected_mask, o, m)
	}
}

func TestDenseIntSet_offset(t *testing.T) {
	assertOffset(t, 0, 0, 0x01)
	assertOffset(t, 1, 0, 0x02)
	for i := uint(0); i < 32; i++ {
		assertOffset(t, i, 0, 1<<i)
	}
	for i := uint(32); i < 64; i++ {
		assertOffset(t, i, 1, 1<<(i-32))
	}
}

func TestDenseIntSet_ensureBits(t *testing.T) {
	set := DenseIntSet{}
	if len(set.bits) != 0 {
		t.Fatalf("Expected set to have length 0, found %d", len(set.bits))
	}
	set.ensureBits(0)
	if len(set.bits) != 1 {
		t.Fatalf("Expected set to have length 1, found %d", len(set.bits))
	}
	if set.bits[0] != 0 {
		t.Fatalf("Expected set to have all 0 bytes, found %v", set.bits)
	}
	set.bits[0] = 10
	// This time should be a no-op
	set.ensureBits(0)
	if len(set.bits) != 1 {
		t.Fatalf("Expected set to have length 1, found %d", len(set.bits))
	}
	if set.bits[0] != 10 {
		t.Fatalf("Expected set to have 10 at byte 0, found %v", set.bits)
	}
	set.ensureBits(1)
	if len(set.bits) != 2 {
		t.Fatalf("Expected set to have length 2, found %d", len(set.bits))
	}
	if set.bits[0] != 10 || set.bits[1] != 0 {
		t.Fatalf("Expected set to have [10, 0], found %v", set.bits)
	}
}

func assertContains(t *testing.T, set DenseIntSet, value uint) {
	if !set.Contains(value) {
		t.Errorf("set was supposed to contain %d", value)
	}
}

func assertNotContains(t *testing.T, set DenseIntSet, value uint) {
	if set.Contains(value) {
		t.Errorf("set was not supposed to contain %d", value)
	}
}

func TestDenseIntSet_AddContainsRemove(t *testing.T) {
	set := DenseIntSet{}
	assertNotContains(t, set, 1)
	assertNotContains(t, set, 10)
	assertNotContains(t, set, 12)
	assertNotContains(t, set, 1000)
	assertNotContains(t, set, 10000)
	set.Add(10)
	assertContains(t, set, 10)
	assertNotContains(t, set, 12)
	set.Add(12)
	assertContains(t, set, 10)
	assertContains(t, set, 12)
	set.Remove(12)
	assertContains(t, set, 10)
	assertNotContains(t, set, 12)
}

func TestDenseIntSet_Add(t *testing.T) {
	set := DenseIntSet{}
	set.Add(10)
	if len(set.bits) != 1 || set.bits[0] != 0x400 {
		t.Errorf("expected set to have bits [0x400] found: %v", set.bits)
	}
	set.Add(64)
	if len(set.bits) != 3 || set.bits[0] != 0x400 || set.bits[1] != 0 || set.bits[2] != 0x01 {
		t.Errorf("expected set to have bits [0x400, 0, 0x01] found: %v", set.bits)
	}
}

func TestDenseIntSet_Remove(t *testing.T) {
	set := DenseIntSet{[]uint32{0x5}}
	if set.Remove(10) {
		t.Errorf("set was not supposed to contain 10 to start with")
	}
	if set.Remove(1000) {
		t.Errorf("set was not supposed to contain 1000 to start with")
	}
	if len(set.bits) != 1 {
		t.Errorf("set bits vector changed size: %d is not length 1", len(set.bits))
	}
	if !set.Remove(0) {
		t.Errorf("Expected set.Remove to indicate that 1 was actually removed")
	}
	if len(set.bits) != 1 || set.bits[0] != 0x04 {
		t.Errorf("Expected set to have bits [0x04] not %v", set.bits)
	}
}

func TestDenseIntSet_Contains(t *testing.T) {
	set := DenseIntSet{[]uint32{0x5}}
	assertContains(t, set, 0)
	assertNotContains(t, set, 1)
	assertContains(t, set, 2)
	assertNotContains(t, set, 3)
	assertNotContains(t, set, 4)
	assertNotContains(t, set, 5)
	assertNotContains(t, set, 6)
	assertNotContains(t, set, 7)
	assertNotContains(t, set, 500)
}
